# data parsing and calculation/__init__.py

"""Retrieve GPS, Seismological, hydrological and gas data.

Modules exported by this package:

- `gps_read`: Provide several gps processing calculations.
- `sil_read`: Provide several seismological processing calculations, useful to correlate with gps data.
"""

